/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file applications/afs/database/contentdatabase.c
 * @author Christian Grothoff
 * @author Igor Wronsky
 *
 * ContentDatabase (dispatch implementation). 
 *
 * This is a wrapper for accessing dynamically
 * loadable actual database routines. Wrapper uses 
 * multiple databases and deterministically selects a bucket 
 * (database) for given content based on its hash. This mechanism 
 * is to overcome the 2G limits in traditional filesystems and 
 * simple database managers.
 *
 **/

#include "config.h"

#include <stdio.h>
#include <string.h>

#include "applications/afs/database/contentdatabase.h"
#include "util/configuration.h"
#include "util/dso.h"
#include "util/storage.h"
#include "util/xmalloc.h"

#define DSO_PREFIX "libgnunetafs_database_"

/**
 * Initialize content database
 * 
 * @param libname name of the dynamic library to use
 * @param databases number of database buckets to use
 * @return handle to the opened database
 **/
DatabaseHandle * initContentDatabase(unsigned char * libname, int databases) {
  DatabaseHandle * result;
  char * dir;
  char * bucketname;
  int i;

  result = MALLOC(sizeof(DatabaseHandle));

  result->library = loadDynamicLibrary(DSO_PREFIX,
  				       libname);
  result->libInitContentDatabase = bindDynamicMethod(result->library,
  				"",
				"libInitContentDatabase");
  result->libDoneContentDatabase = bindDynamicMethod(result->library,
  				"",
				"libDoneContentDatabase");
  result->libUnlinkFromDB        = bindDynamicMethod(result->library,
  				"",
				"libUnlinkFromDB");
  result->libCountContentEntries = bindDynamicMethod(result->library,
  				"",
				"libCountContentEntries");
  result->libReadContent         = bindDynamicMethod(result->library,
  				"",
				"libReadContent");
  result->libWriteContent        = bindDynamicMethod(result->library,
  				"",
				"libWriteContent");
  result->libForEachEntryInDatabase = bindDynamicMethod(result->library,
  				"",
				"libForEachEntryInDatabase");
  result->databaseCount = databases;
  
  dir = getFileName("AFS",
		    "CONTENTDIR",
		    "Configuration file must specify directory for the content database in section %s under %s.\n");
  mkdirp(dir);

  bucketname = MALLOC(strlen(dir) + strlen("bucket") + 12);
  result->dbfs = MALLOC(result->databaseCount*sizeof(void *));

  for(i=0;i<result->databaseCount;i++) {
    sprintf(bucketname, "%s/bucket.%d", dir, i);
    result->dbfs[i]=result->libInitContentDatabase(bucketname);
  }
  FREE(bucketname);
  FREE(dir);

  return result;
}

/**
 * Clean shutdown of the storage module
 **/
void doneContentDatabase(DatabaseHandle * handle) {
  int i;

  for(i=0;i<handle->databaseCount;i++) {
    handle->libDoneContentDatabase(handle->dbfs[i]);
  }
  FREE(handle->dbfs);
  unloadDynamicLibrary(handle->library);
  FREE(handle);
}

/**
 * Call a method for each key in the database and
 * call the callback method on it. 
 *
 * @param callback the callback method
 * @param data second argument to all callback calls
 * @return the number of items stored in the content database
 **/
int forEachEntryInDatabase(DatabaseHandle * handle,
			   EntryCallback callback,
			   void * data){
  int total=0;
  int i;

  for(i=0;i<handle->databaseCount;i++) {
    total += handle->libForEachEntryInDatabase(handle->dbfs[i], callback, data);
  }
  return total;
}

/**
 * Get the number of entries in the database.
 * @return SYSERR on error, otherwise the number of entries
 **/
int countContentEntries(DatabaseHandle * handle) {
  int i;
  int total=0;
  
  for(i=0;i<handle->databaseCount;i++) {
    total += handle->libCountContentEntries(handle->dbfs[i]);
  }
  return total;
}
 
/**
 * Read the contents of a bucket to a buffer.
 *
 * @param fn the hashcode representing the entry
 * @param result the buffer to write the result to 
 *        (*result should be NULL, sufficient space is allocated)
 * @return the number of bytes read on success, -1 on failure
 **/ 
int readContent(DatabaseHandle * handle,
		HashCode160 * name,
		void ** result) {
  int bucket = (name->c % handle->databaseCount);
  
  return handle->libReadContent(handle->dbfs[bucket], name, result);
}

/**
 * Write content to a file. Check for reduncancy and eventually
 * append.
 *
 * @param fn the key for the entry
 * @param len the size of the block
 * @param block the data to store
 * @return SYSERR on error, OK if ok.
 **/
int writeContent(DatabaseHandle * handle,
		 HashCode160 * fn, 
		 int len,
		 void * block) {
  int bucket = (fn->c % handle->databaseCount);
  
  return handle->libWriteContent(handle->dbfs[bucket], fn, len, block);
}

/**
 * Free space in the database by removing an entry.
 *
 * @param fn the key of the entry to remove
 * @return SYSERR on error, OK if ok.
 **/
int unlinkFromDB(DatabaseHandle * handle,
		 HashCode160 * name) {
  int bucket = (name->c % handle->databaseCount);
  
  return handle->libUnlinkFromDB(handle->dbfs[bucket], name);
}


/* end of contentdatabase.c */
