/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * ContentDatabase (directory based implementation). 
 * @author Christian Grothoff
 * @file applications/afs/database/contentdatabase_directory.c
 **/

#include "config.h"

#include <sys/stat.h>

#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "util/logging.h"
#include "util/returnvalues.h"
#include "util/statuscalls.h"
#include "util/storage.h"
#include "util/xmalloc.h"
#include "applications/afs/database/contentdatabase.h"

#define DIRECTORY_DEBUG 0

#define DIR_EXT ".dir"

/**
 * Initialize the Directory module, expand filename
 * @param dir the directory where content is configured to be stored (e.g. ~/.gnunet/data/content).
 **/
static char * getDirectory(char * dir) {
  char * result;
  char * tmp;

#if DIRECTORY_DEBUG
  LOG(LOG_INFO, 
      "Database (Directory): %s\n", 
      dir);
#endif
  tmp = MALLOC(strlen(dir) + strlen(DIR_EXT) + 5);
  sprintf(tmp, "%s%s/", dir, DIR_EXT);
  result = expandFileName(tmp);
  FREE(tmp);
  return result;
}

void * libInitContentDatabase(char * dir) {  
  char * dbh;
  
  dbh = getDirectory(dir);  
  if (dbh == NULL) 
    errexit("FATAL: could not open directory %s!\n",
	    (char*) dbh);
  mkdirp(dbh);
  registerSpaceMonitor(dbh); 
  return dbh;
}

/**
 * Clean shutdown of the storage module (not used at the moment)
 **/
void libDoneContentDatabase(char * dbh) {
  FREE(dbh);
}

/**
 * Call a method for each entry in the database and
 * call the callback method on it. 
 * @return the number of items stored in the content database
 **/
int libForEachEntryInDatabase(char * dbh,
			      EntryCallback callback,
			      void * data) {
  DIR * dinfo;
  struct dirent *finfo;
  struct stat istat;
  int count;
  HashCode160 hash;

  stat(dbh, &istat);
  if (!S_ISDIR(istat.st_mode)) {
    LOG(LOG_ERROR,
	"ERROR: content database directory %s is not a directory.\n",
	dbh);
    return -1;
  }
  errno = 0;
  dinfo = opendir(dbh);
  if ((errno == EACCES) || (dinfo == NULL)) {
    LOG(LOG_ERROR,
	"ERROR: access to %s was denied (%s)\n",
	dbh, 
	strerror(errno));
    return -1;
  }
  count = 0;
  while ((finfo = readdir(dinfo)) != NULL)
    if (strlen(finfo->d_name) == sizeof(HashCode160)*2) {
      if (callback != NULL) {
	hex2hash((HexName*) finfo->d_name,
		 &hash);
	callback(&hash, data);
      }
      count++;
    }

  closedir(dinfo);
  return count;
}

/**
 * How many entries are in the database?
 * @return the number of entries, -1 on failure
 **/ 
int libCountContentEntries(char * dbh) {
  return libForEachEntryInDatabase(dbh, NULL, NULL);
}

/**
 * Read the contents of a bucket to a buffer.
 *
 * @param fn the hashcode representing the entry
 * @param result the buffer to write the result to 
 *        (*result should be NULL, sufficient space is allocated)
 * @return the number of bytes read on success, -1 on failure
 **/ 
int libReadContent(char * dbh,
		   HashCode160 * name,
	           void ** result) {
  /* open file, must exist, open read only */
  int handle;
  int size;
  HexName fn;
  char * fil;
  size_t fsize;

  if ((name == NULL) || (result == NULL))
    return -1;
  hash2hex(name, &fn);
  fil = MALLOC(strlen(dbh) + strlen((char*)&fn) + 1);
  buildFileName(dbh, &fn, fil);
  handle = open(fil, O_RDONLY, S_IRUSR);
  fsize = getFileSize(fil);
  FREE(fil);

  if (handle == -1)
    return -1;
  *result = MALLOC(fsize);
  size = read(handle, *result, fsize);
  close(handle);
  if (size == -1) {
    FREE(*result);
    *result = NULL;
  }
  return size;
}


/**
 * Write content to a file. Check for reduncancy and eventually
 * append.
 *
 * @param name the key for the entry
 * @param len the size of the block
 * @param block the data to store
 * @return SYSERR on error, OK if ok.
 **/
int libWriteContent(char * dbh,
		    HashCode160 * name, 
		    int len,
		    void * block) {
  HexName fn;
  char * fil;
  int handle;

  hash2hex(name, &fn);
  fil = MALLOC(strlen(dbh) + strlen((char*)&fn) + 1);
  buildFileName(dbh, &fn, fil);
  handle = open(fil, O_RDWR|O_CREAT, S_IRUSR|S_IWUSR);
  if (handle == -1) {
    LOG(LOG_WARNING,
	"WARNING: Failed to open file %s\n", fil);
    FREE(fil);
    return SYSERR; /* failed! */
  }
  FREE(fil);
  write(handle, block, len);
  close(handle);
  increaseKilobyteUsed(1);
  return OK;
}

/**
 * Free space in the database by removing one file
 * @param name the hashcode representing the name of the file 
 *        (without directory)
 **/
int libUnlinkFromDB(char * dbh,
		    HashCode160 * name) {
  HexName fn;
  char * fil;

  if (name == NULL)
    return SYSERR;
  hash2hex(name, &fn);
  fil = MALLOC(strlen(dbh) + strlen((char*)&fn) + 1);
  buildFileName(dbh, &fn, fil);
  unlink(fil);
  FREE(fil);
  return OK;
}

/* end of contentdatabase_directory.c */
