/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * ContentDatabase (tdb implementation). 
 * @author Christian Grothoff
 * @author Uli Luckas
 * @author Igor Wronsky
 * @file applications/afs/database/contentdatabase_tdb.c
 **/

#include "config.h"

#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

#include "applications/afs/database/contentdatabase_tdb.h"
#include "util/xmalloc.h"
#include "util/logging.h"
#include "util/storage.h"
#include "util/returnvalues.h"
#include "util/statuscalls.h"

#define TDB_DEBUG 0

#define GIGA_BYTE (1024 * 1024 * 1024)

/**
 * Extension for the TDB database.
 **/
#define TDB_EXT ".tdb"

/**
 * After how-many insert operations test
 * DB size?
 **/
#define TEST_FREQUENCY 1024

/**
 * Open a tdb database (for content)
 * @param dir the directory where content is configured 
 * to be stored (e.g. data/content). A file called ${dir}.dbf is used instead
 **/
static tdbHandle * getDatabase(char * dir) {
  char * ff;
  tdbHandle * dbh;
 
  int fnSize = strlen(dir) + strlen(TDB_EXT) + 1;

#if TDB_DEBUG  
  LOG(LOG_DEBUG,
      "DEBUG: Database: %s (TDB)\n", 
      dir);
#endif
  dbh = MALLOC(sizeof(tdbHandle));
  ff = MALLOC(fnSize);
  strcpy(ff, dir);
  if (ff[strlen(ff)-1] == '/')
    ff[strlen(ff) - 1] = 0; /* eat the '/' at the end */
  else
    ff[strlen(ff)] = 0; /* no '/' to eat */
  strcat(ff, TDB_EXT);
  dbh->filename = expandFileName(ff);
 
  dbh->dbf = tdb_open(dbh->filename, 0, TDB_NOMMAP, O_RDWR | O_CREAT, S_IRUSR|S_IWUSR);
  if (NULL == dbh->dbf) 
    errexit("getDatabase: failed to open database file %s with error: %s\n",
	    dbh->filename,
	    tdb_errorstr(dbh->dbf));
  registerSpaceMonitor(dbh->filename);

  FREE(ff);
  dbh->insertCount = TEST_FREQUENCY; 
  return dbh;
}

void * libInitContentDatabase(char * dir) {
  tdbHandle * dbh;
  
  dbh = getDatabase(dir);
  MUTEX_CREATE_RECURSIVE(&dbh->DATABASE_Lock_);	

  return dbh;
}

/**
 * Clean shutdown of the storage module (not used at the moment)
 **/
void libDoneContentDatabase(tdbHandle * dbh) {
 
  MUTEX_DESTROY(&dbh->DATABASE_Lock_);
  tdb_close(dbh->dbf);
  FREE(dbh->filename);
  FREE(dbh);
}


/**
 * Call a method for each key in the database and
 * call the callback method on it. 
 *
 * @param callback the callback method
 * @param data second argument to all callback calls
 * @return the number of items stored in the content database
 **/
int libForEachEntryInDatabase(tdbHandle * dbh,
 		           EntryCallback callback,
	    	           void * data) {
  TDB_DATA prevkey, key;
  HashCode160 doubleHash;
  int count;

  count = 0;
  /* scan database data/content.dbf and add entries to database 
     if not already present */
  MUTEX_LOCK(&dbh->DATABASE_Lock_);
  key = tdb_firstkey(dbh->dbf);
  while (key.dptr) {
    if (strlen(key.dptr) == sizeof(HashCode160)*2) {	
      if (callback != NULL) {
	  hex2hash((HexName*)key.dptr,
   	    &doubleHash);
  	  callback(&doubleHash, 
		   data);
      }
      count++; /* one more file */
    }
    prevkey = key;
    key = tdb_nextkey(dbh->dbf, prevkey);
    FREE(prevkey.dptr);
  }
  MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
  return count;
}

#define COUNTENTRY "count_token"

static void storeCount(tdbHandle * dbh,
		       int count) {
  TDB_DATA key;
  TDB_DATA buffer;

  key.dptr = COUNTENTRY;
  key.dsize = strlen(COUNTENTRY)+1;
  buffer.dptr = (char*)&count;
  buffer.dsize = sizeof(int);
  MUTEX_LOCK(&dbh->DATABASE_Lock_);
  tdb_store(dbh->dbf, key, buffer, TDB_REPLACE);
  MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
}

/**
 * Get the number of entries in the database.
 **/
int libCountContentEntries(tdbHandle * dbh) {
  TDB_DATA key;
  TDB_DATA buffer;
  int count;

  key.dptr = COUNTENTRY;
  key.dsize = strlen(COUNTENTRY)+1;

  MUTEX_LOCK(&dbh->DATABASE_Lock_);
  buffer = tdb_fetch(dbh->dbf, key);
  MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
  if ( (!buffer.dptr) || 
       (buffer.dsize != sizeof(int)) ) {
    count = libForEachEntryInDatabase(dbh, NULL, NULL);  
    storeCount(dbh, count);
  } else {
    count = *(int*) buffer.dptr;
    free(buffer.dptr);
  }
  return count;
}

/**
 * Read the contents of a bucket to a buffer.
 *
 * @param fn the hashcode representing the entry
 * @param result the buffer to write the result to 
 *        (*result should be NULL, sufficient space is allocated)
 * @return the number of bytes read on success, -1 on failure
 **/ 
int libReadContent(tdbHandle * dbh,
	        HashCode160 * name,
	        void ** result) {
  HexName fn;
  TDB_DATA key, buffer;

  hash2hex(name, &fn);  
  key.dptr = fn.data;
  key.dsize = strlen(key.dptr) + 1;

  MUTEX_LOCK(&dbh->DATABASE_Lock_);
  buffer = tdb_fetch(dbh->dbf, key);
  MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
  if (!buffer.dptr) 
    return -1;
  *result = buffer.dptr;
  return buffer.dsize;  
}

/**
 * Write content to a file. Check for reduncancy and eventually
 * append.
 *
 * @param fn the key for the entry
 * @param len the size of the block
 * @param block the data to store
 * @return SYSERR on error, OK if ok.
 **/
int libWriteContent(tdbHandle * dbh,
		 HashCode160 * name, 
		 int len,
		 void * block) {
  HexName fn;
  TDB_DATA buffer, key;

  if (dbh->insertCount > 1024) {
    if (getFileSize(dbh->filename) > 
	(unsigned long long) 2 * GIGA_BYTE - 3 * TEST_FREQUENCY * len) {
      LOG(LOG_WARNING,
	  "WARNING: single tdb database is limited to 2 GB, can not store more data.\n");
      return SYSERR; /* enforce TDB size limit of 2 GB minus 3*TF_len slack */
    }
    dbh->insertCount = 0;
  } else
    dbh->insertCount++;
 
  hash2hex(name, &fn);
  key.dptr = fn.data;
  key.dsize = strlen(key.dptr) + 1;
  buffer.dptr = block;
  buffer.dsize = len;
  MUTEX_LOCK(&dbh->DATABASE_Lock_);
  tdb_store(dbh->dbf, key, buffer, TDB_REPLACE);
  MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
  storeCount(dbh, libCountContentEntries(dbh) + 1);
  increaseKilobyteUsed(1);
  return OK;
}

/**
 * Free space in the database by removing one file
 * @param name hashcode representing the name of the file (without directory)
 **/
int libUnlinkFromDB(tdbHandle * dbh,
		 HashCode160 * name) {
  TDB_DATA key;
  HexName fn;

  hash2hex(name, &fn);
 
  key.dptr = fn.data;
  key.dsize = strlen(key.dptr) + 1;
  MUTEX_LOCK(&dbh->DATABASE_Lock_);
  tdb_delete(dbh->dbf, key);
  MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
  storeCount(dbh, libCountContentEntries(dbh) - 1);
  return OK;
}

/* end of contentdatabase_tdb.c */
