/*
     This file is part of GNUnet

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file src/applications/afs/gtkui/search.c
 * @author Christian Grothoff
 * @author Igor Wronsky
 *
 * This handles the box displaying search results for
 * the gtk+ client.
 **/

#include "config.h"

#include <errno.h>
#include <gtk/gtk.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "util/configuration.h"
#include "util/logging.h"
#include "util/returnvalues.h"
#include "util/xmalloc.h"
#include "applications/afs/gtkui/helper.h"
#include "applications/afs/gtkui/download.h"
#include "applications/afs/gtkui/saveas.h"
#include "util/port.h"

/**
 * This method is called whenever the user clicks the
 * download button. 
 *
 * @param listModel Data related to the search
 **/
static void download(GtkWidget * widget,
		     ListModel * listModel) {  
  DownloadModel * dlm;
  RootNode * root;
  int descLen, nameLen;
  gchar * tmp;
  gint row;
  
  row = listModel->selectedRow;
  if (row == -1) {
    guiMessage("Nothing selected!\n");
    return;
  }

  dlm = MALLOC(sizeof(DownloadModel));
  memset(dlm, 
  	 0,
	 sizeof(DownloadModel));
  /* Collect data required from clist entry */
  root = &dlm->root;
  gtk_clist_get_text(GTK_CLIST(listModel->search_result_list),
 		     row, 1, &tmp);
  sscanf(tmp, "%d", &root->header.fileIdentifier.file_length);
  gtk_clist_get_text(GTK_CLIST(listModel->search_result_list),
 		     row, 3, &tmp);
  sscanf(tmp, "%d", &root->header.fileIdentifier.crc);
  gtk_clist_get_text(GTK_CLIST(listModel->search_result_list),
 		     row, 4, &tmp);
  hex2hash((HexName*)tmp,
  	   &root->header.fileIdentifier.chk.query);
  gtk_clist_get_text(GTK_CLIST(listModel->search_result_list),
 		     row, 5, &tmp);
  hex2hash((HexName*)tmp,
  	   &root->header.fileIdentifier.chk.key);
  gtk_clist_get_text(GTK_CLIST(listModel->search_result_list),
 		     row, 0, &tmp);
  descLen=strlen(tmp);
  if(descLen >= MAX_DESC_LEN) {
    LOG(LOG_WARNING,
        "WARNING: Description too long? (%s)",
	tmp);
    descLen=MAX_DESC_LEN-1;
  }
  memcpy(&root->header.description,
  	 tmp,
	 descLen+1);
  root->header.description[descLen]=0;
  gtk_clist_get_text(GTK_CLIST(listModel->search_result_list),
 		     row, 2, &tmp);
  nameLen=strlen(tmp);
  if(nameLen >= MAX_FILENAME_LEN) {
    nameLen = MAX_FILENAME_LEN-1;
  }
  memcpy(root->header.filename,
  	 tmp,
	 nameLen);
  root->header.filename[nameLen]=0;
  dlm->listModel = listModel;

  gtk_clist_freeze(GTK_CLIST(listModel->search_result_list));
  
  /* Open the save-as -window */
  openSaveAs(dlm);
}

/**
 * User has selected some search list row
 **/
static void select_row_callback(GtkWidget *widget,
				gint row,
				gint column,
				GdkEventButton *event,
				gpointer data) {
  if(GTK_CLIST(((ListModel*)data)->search_result_list)->freeze_count>0)
    return;
  ((ListModel*)data)->selectedRow = row;
}

/**
 * Some search list row has been unselected
 **/
static void unselect_row_callback(GtkWidget *widget,
				  gint row,
				  gint column,
				  GdkEventButton *event,
				  gpointer data) {
  if(GTK_CLIST(((ListModel*)data)->search_result_list)->freeze_count>0)
    return;
  ((ListModel*)data)->selectedRow = -1;
}

/**
 * Display results!
 *
 * @param rootNode Data about a file
 * @param model Data related to the search
 **/
static void displayResult(RootNode * rootNode,
			  ListModel * model) {
  GtkWidget * search_result_list;
  gchar ** results;

  rootNode->header.description[MAX_DESC_LEN-1] = 0;
  rootNode->header.filename[MAX_FILENAME_LEN-1] = 0;
  rootNode->header.mimetype[MAX_MIMETYPE_LEN-1] = 0;

  search_result_list = model->search_result_list;
  results = MALLOC(7*sizeof(gchar *));
  results[0] = rootNode->header.description;
  results[1] = MALLOC(32);
  sprintf(results[1],"%d", 
	  ntohl(rootNode->header.fileIdentifier.file_length));
  results[2] = rootNode->header.filename;
  results[3] = MALLOC(32);
  sprintf(results[3],"%d", 
	  ntohl(rootNode->header.fileIdentifier.crc));
  results[4] = MALLOC(sizeof(HexName));
  hash2hex(&rootNode->header.fileIdentifier.chk.query,
	   (HexName*) results[4]);
  results[5] = MALLOC(sizeof(HexName));
  hash2hex(&rootNode->header.fileIdentifier.chk.key,
           (HexName*) results[5]);
  results[6] = rootNode->header.mimetype;
  
  gdk_threads_enter();
  gtk_clist_freeze(GTK_CLIST(search_result_list));
  gtk_clist_append(GTK_CLIST(search_result_list),
		   results);
  gtk_clist_thaw(GTK_CLIST(search_result_list));
  gdk_threads_leave();

  FREE(results[1]);
  FREE(results[3]);
  FREE(results[4]);
  FREE(results[5]);
  FREE(results);
}

typedef struct {
  int num_keywords;
  HashCode160 * keys;
  AFS_CS_QUERY ** messages;
  void * displayResult;
  ListModel * model;
} _receiveResultArgs_;

/**
 * Should the receive thread abort? (has the window been closed?)
 *
 * @return YES if we should abort
 **/
static int testTermination(ListModel * model) {
  if (model->sem == NULL)
    return NO;
  else
    return YES;
}

/**
 * Main method of the receiveResults-threads...
 **/
static void receiveResults_(_receiveResultArgs_ * args) {
  receiveResults(args->model->SEARCH_socket_,	
  		 args->num_keywords,
		 args->keys,
		 args->messages,
		 args->displayResult,
		 args->model,
		 (TestTerminateThread)&testTermination,
		 args->model);
  SEMAPHORE_UP(args->model->sem); /* signal: thread terminated */
  FREE(args->model->sqc);
  FREE(args->messages);
  FREE(args->keys);
  FREE(args);
}

/**
 * The main method of the search-thread.
 *
 * @param searchString What to look for
 * @param model Data related to the search
 **/
void startSearchThreads(char * searchString,
			ListModel * model) {
  HashCode160 * keys;
  int num_keywords;
  AFS_CS_QUERY ** messages;
  _receiveResultArgs_ * receiveArgs;
  SendQueriesContext * sqc;
  pthread_t receiveThread;

  /*  fprintf(stderr,"Searching for %s!\n",
      searchString);*/

  num_keywords = parseKeywords(1,
			       &searchString,
			       &keys);
  if (num_keywords <= 0) 
    return;
  
  buildMessages(num_keywords,
		keys,
		&messages);
  sqc = MALLOC(sizeof(SendQueriesContext));
  sqc->messages = messages;
  sqc->queryCount = num_keywords;
  sqc->sock = model->SEARCH_socket_;
  sqc->timeout = getConfigurationInt("AFS",
				     "SEARCHTIMEOUT") * cronSECONDS; 
  cronTime(&sqc->start);
  model->sqc = sqc;
  addCronJob((CronJob)&sendQueries,
	     0, 
	     0,
	     sqc);

  receiveArgs = MALLOC(sizeof(_receiveResultArgs_));
  receiveArgs->num_keywords = num_keywords;
  receiveArgs->keys = keys;
  receiveArgs->messages = messages;
  receiveArgs->displayResult = displayResult;
  receiveArgs->model = model;  
  if (0 != PTHREAD_CREATE(&receiveThread,
			  (void * (*)(void *)) &receiveResults_,
			  receiveArgs,
			  16 * 1024)) {
    errexit("FATAL: could not create receive thread (%s)!\n",
	    strerror(errno));    
  } else
    pthread_detach(receiveThread);
}

/**
 * Stop the search thread and free the model. This method MUST always
 * be called when the search ends, either because the search was
 * aborted or because gnunet-gtk exists as a whole.
 *
 * @param widget the window (not used)
 * @param model the model with the socket and the sqc 
 **/
static void stopSearch(GtkWidget * widget,
		       ListModel * model) { 
  LOG(LOG_DEBUG, 
      "DEBUG: stopSearch called\n");
  model->sem = SEMAPHORE_NEW(0); 
  destroySocket(model->SEARCH_socket_);
  delCronJob((CronJob)&sendQueries, 
	     0, 
	     model->sqc);
  SEMAPHORE_DOWN(model->sem); /* see receiveResults_ */
  SEMAPHORE_FREE(model->sem);
  releaseClientSocket(model->SEARCH_socket_); /* now free memory of socket, too. */
  FREE(model);
}

/**
 * A function for case-insensitive text comparisons
 **/
static gint string_casecmp(GtkCList *clist,
			   gconstpointer ptr1,
			   gconstpointer ptr2) {
  char * text1;
  char * text2;
  GtkCListRow * row1 = (GtkCListRow *) ptr1;
  GtkCListRow * row2 = (GtkCListRow *) ptr2;

  text1 = GTK_CELL_TEXT (row1->cell[clist->sort_column])->text;
  text2 = GTK_CELL_TEXT (row2->cell[clist->sort_column])->text;
 
  return (strcasecmp(text1,text2));	
}

/**
 * A function for numeric comparisons
 **/
static gint string_numericcmp(GtkCList *clist, 
			      gconstpointer ptr1,
			      gconstpointer ptr2) {
  int value1;
  int value2;
  GtkCListRow * row1 = (GtkCListRow *) ptr1;
  GtkCListRow * row2 = (GtkCListRow *) ptr2;
  
  value1 = atoi(GTK_CELL_TEXT (row1->cell[clist->sort_column])->text);
  value2 = atoi(GTK_CELL_TEXT (row2->cell[clist->sort_column])->text);
  
  if(value1>value2)
    return(-1);
  else if(value1==value2)
    return(0);
  else
    return(1);
}

/**
 * Changes the current sort column and sorts the list.
 **/
void sort_column_callback(GtkCList * clist,
			  gint column,
			  gpointer data) {
  /* Sort all columns as strings except 1 */
  if(column==1)
    gtk_clist_set_compare_func(clist, (GtkCListCompareFunc)string_numericcmp);
  else
    gtk_clist_set_compare_func(clist, (GtkCListCompareFunc)string_casecmp);
  
  gtk_clist_set_sort_column(clist, column);
  gtk_clist_freeze(clist);
  gtk_clist_sort(clist);
  gtk_clist_thaw(clist);
}

/** 
 * Returns a box containing the search results list.
 **/
GtkWidget * getSearchWindow(gchar * title) {
  GtkWidget * scrolled_window;
  GtkWidget * button;
  GtkWidget * box;
  GtkWidget * search_result_list;
  ListModel * model;
  static gchar * descriptions[] = {
    "Description",
    "Size",
    "Filename",
    "CRC",
    "Key1",
    "Key2",
    "Mimetype"};

  model = (ListModel*) MALLOC(sizeof(ListModel));
  model->selectedRow = -1;
  model->sem = NULL;
  box = gtk_vbox_new(FALSE, 0);
  
  /* scrolled window for the list */
  scrolled_window = gtk_scrolled_window_new (NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW (scrolled_window),
                                 GTK_POLICY_AUTOMATIC, 
				 GTK_POLICY_ALWAYS);
  gtk_box_pack_start(GTK_BOX(box), 
		     scrolled_window, TRUE, TRUE, 0);
  gtk_widget_show(scrolled_window);
  
  /* result set list */
  search_result_list 
    = gtk_clist_new_with_titles(7, descriptions);
  model->search_result_list = search_result_list;
  model->SEARCH_socket_ = getClientSocket();

  gtk_container_add(GTK_CONTAINER(scrolled_window), search_result_list);
  gtk_widget_show(search_result_list);

  gtk_clist_set_selection_mode
    (GTK_CLIST(search_result_list),
     GTK_SELECTION_SINGLE);
  /* set passive titles as default */
  gtk_clist_column_titles_passive
    (GTK_CLIST(search_result_list));

  /* allow sorting by description, size, filename and mimetype */
  gtk_clist_column_title_active(GTK_CLIST(search_result_list),
    0);
  gtk_clist_column_title_active(GTK_CLIST(search_result_list),
    1);
  gtk_clist_column_title_active(GTK_CLIST(search_result_list),
    2);
  gtk_clist_column_title_active(GTK_CLIST(search_result_list),
    6);
  gtk_signal_connect(GTK_OBJECT(search_result_list),
		     "click-column",
		     GTK_SIGNAL_FUNC(sort_column_callback),
		     NULL);

  /* description left, size right justification */
  gtk_clist_set_column_justification
    (GTK_CLIST(search_result_list),
     0,
     GTK_JUSTIFY_LEFT);
  gtk_clist_set_column_justification
    (GTK_CLIST(search_result_list),
     1,
     GTK_JUSTIFY_RIGHT);
  gtk_signal_connect(GTK_OBJECT(search_result_list),
		     "unselect_row",
		     GTK_SIGNAL_FUNC(unselect_row_callback),
		     (gpointer) model);
  gtk_signal_connect(GTK_OBJECT(search_result_list),
		     "select_row",
		     GTK_SIGNAL_FUNC(select_row_callback),
		     (gpointer) model);

  /* What however is important, is that we set the column widths as
   * they will never be right otherwise. Note that the columns are
   * numbered from 0 and up (to 6 in this case).
   */
  gtk_clist_set_column_width (GTK_CLIST(search_result_list), 
			      0, 490);
  gtk_clist_set_column_width (GTK_CLIST(search_result_list), 
			      1, 70);
  gtk_clist_set_column_width (GTK_CLIST(search_result_list), 
			      2, 200);
  gtk_clist_set_column_width (GTK_CLIST(search_result_list), 
			      3, 40);
  gtk_clist_set_column_width (GTK_CLIST(search_result_list), 
			      4, 40);
  gtk_clist_set_column_width (GTK_CLIST(search_result_list), 
			      5, 40);
  gtk_clist_set_column_width (GTK_CLIST(search_result_list), 
			      6, 200);

  /* download button */
  button = gtk_button_new_with_label("Download");
  gtk_signal_connect (GTK_OBJECT (button), 
		      "clicked",
		      GTK_SIGNAL_FUNC (download), 
		      model);
  gtk_box_pack_start(GTK_BOX(box), button, FALSE, FALSE, 0);
  gtk_widget_show(button);
  /* display window */
  startSearchThreads(title,
		     model);
  gtk_signal_connect(GTK_OBJECT(scrolled_window), 
                     "delete_event",
		     GTK_SIGNAL_FUNC(deleteEvent), 
		     NULL);
  gtk_signal_connect(GTK_OBJECT(scrolled_window), 
                     "destroy",
		     GTK_SIGNAL_FUNC(stopSearch), 
		     model);
  return (box);
}

/* end of search.c */
