/*
     This file is part of GNUnet

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/
/**
 * ContentDatabse - block storage using a GDBM database.
 * @author Christian Grothoff
 * @author Igor Wronsky
 * @file include/applications/afs/database/contentdatabase_gdbm.h
 **/

#ifndef CONTENTDATABASE_GDBM_H
#define CONTENTDATABASE_GDBM_H

#include "config.h"

#include <gdbm.h>

#include "util/semaphore.h"
#include "applications/afs/database/contentdatabase.h"

/**
 * A gdbm wrapper
 **/
typedef struct {

  /**
   * GDBM handle.
   **/
  GDBM_FILE dbf;

  /**
   * Name of the database file (for size-tests).
   **/
  char * filename;

  /**
   * How many blocks have we deleted so far? After
   * deleting about a thousand we should call reorganize
   * to ensure that we actually reclaim the space...
   **/
  int delCount;

  /**
   * Number of insert operations since last 
   * size-check?
   **/
  int insertCount;

  /**
   * gdbm requires synchronized access.
   **/
  Mutex DATABASE_Lock_;

} gdbmHandle;

/**
 * Initialize the storage module.
 * @param dir the name of the directory/file
 *        containing the content database
 **/
void * libInitContentDatabase_GDBM(char * dir);

/**
 * Clean shutdown of the storage module.
 **/
void libDoneContentDatabase_GDBM(gdbmHandle * dbh);


/**
 * Free space in the database by removing an entry.
 *
 * @param fn the key of the entry to remove
 * @return SYSERR on error, OK if ok.
 **/
int libUnlinkFromDB(gdbmHandle * dbh, HashCode160 * fn);
 
/**
 * Get the number of entries in the database.
 * @return SYSERR on error, otherwise the number of entries
 **/
int libCountContentEntries(gdbmHandle * dbh);

/**
 * Read the contents of a bucket to a buffer.
 *
 * @param fn the hashcode representing the entry
 * @param result the buffer to write the result to 
 *        (*result should be NULL, sufficient space is allocated)
 * @return the number of bytes read on success, SYSERR on failure
 **/ 
int libReadContent(gdbmHandle * dbh,
		   HashCode160 * fn,
		   void ** result);

/**
 * Write content to a file. Check for reduncancy and eventually
 * append.
 *
 * @param fn the key for the entry
 * @param len the size of the block
 * @param block the data to store
 * @return SYSERR on error, OK if ok.
 **/
int libWriteContent(gdbmHandle * dbh,
		    HashCode160 * fn, 
		    int len,
		    void * block);

/**
 * Call a method for each key in the database and
 * call the callback method on it. 
 *
 * @param callback the callback method
 * @param data second argument to all callback calls
 * @return the number of items stored in the content database
 **/
int libForEachEntryInDatabase(gdbmHandle * dbh,
			      EntryCallback callback,
			      void * data);
 
#endif

/* end of contentdatabase_gdbm.h */
